#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import os

import urwid

from ytsearch import ui
from ytsearch.ui import page


class Interface(page.Page):

    """
    An interface class for the cached items page. This loads information
    from the CACHE_LOCATION you have set.

    :page.Page: The page class to base this one off.
    """

    widgets = None
    results = []

    def __init__(self, parent):
        self.parent = parent

    def load_page(self):
        """Draw the widgets and return the page.
        
        :return: urwid.Pile: The widget to display in the page section.
        """
        title = urwid.Filler(urwid.Text(('title', u'Cached Items')), 'top')
        videos = self.load_videos()
        pile = urwid.Pile([(2, title), ('weight', 1, videos)])
        self.widgets = pile
        return pile

    def load_videos(self):
        """Load all of the videos from your CACHE_LOCATION
        
        :return: urwid.ItemList: A widget with a list of videos.
        """
        videos = []
        self.results = []
        files = {os.path.splitext(v)[0]: v for v in
                 os.listdir(ui.CACHE_LOCATION)}
        for name, fullname in sorted(files.items()):
            fullpath = '{}{}'.format(ui.CACHE_LOCATION, fullname)
            video = ui.Video(name, fullpath, 'file', fullpath)
            self.results.append(video)
            widget = self.parent.create_video_widget(video)
            videos.append(widget)
        self.walker = urwid.SimpleListWalker(videos)
        self.video_list = ui.ItemList(self.parent, self.walker, 'cache')
        return self.video_list



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import urllib.parse
import os
import json
import functools

import requests

from ytsearch import settings, ui


KEY = 'AIzaSyCfkNdqG96tVbzfeybUG9Qk-zQn4txtlWc'
BASE_URL = 'https://www.googleapis.com/youtube/v3/'
CACHE_LOCATION = settings.CACHE_LOCATION


def search(query):
    """
    Searches for a youtube video.
    
    :query: str: The video query to search for.
    :return: json: A result for the youtube api.
             None: If the query failed it returns None.
    """
    params = {'part': 'snippet', 'key': KEY, 'q': urllib.parse.quote(query),
              'maxResults': 50}
    results = request('search', **params)
    if results is None:
        return None
    videos = []
    cache = {os.path.splitext(x)[0]:'{}{}'.format(CACHE_LOCATION, x)
             for x in os.listdir(CACHE_LOCATION)}
    for item in results['items']:
        video_id = item['id'].get('videoId', None)
        if video_id is None:
            continue
        name = item['snippet']['title']
        video = ui.Video(name, video_id, 'youtube', cache.get(name, None))
        videos.append(video)
    return videos


def raw_search(query):
    params = {'part': 'snippet', 'key': KEY, 'q': urllib.parse.quote(query),
              'maxResults': 50}
    return request('search', **params)


def get_playlist(playlist_id):
    """
    Get a youtube playlist

    :playlist_id: str: The ID of the playlist.
    :return: Video: A video instance.
             dict: Information about the playlist.
    """
    current_page = None
    next_page = None
    videos = []
    while current_page != next_page or current_page is None:
        current_page = next_page
        params = {'part': 'snippet,contentDetails', 'playlistId': playlist_id,
                  'key': KEY, 'maxResults': 50, 'pageToken': current_page}
        response = request('playlistItems', **params)
        items = response.get('items', [])
        videos.extend(items)
        next_page = response.get('nextPageToken', None)
        if next_page is None:
            break 
    params = {'part': 'snippet', 'playlistId': playlist_id, 'key': KEY,
              'id': playlist_id}
    name_result = request('playlists', **params)
    name_item = name_result['items'][0]
    return videos, name_item


def request(url, **params):
    """
    Performs a youtube API request.
    
    :url: str: The section of the API to call.
    :params: dict: A dictionary of key / value pairs to use as url params.
    :return: None: Returns None if there was an error.
             json: Returns a json mapping of the API results.
    """
    param_list = ['{}={}'.format(x, params[x]) for x in params
                  if params[x] is not None]
    req = requests.get('{}{}?{}'.format(BASE_URL, url, '&'.join(param_list)))
    try:
        result = req.json()
    except json.decoder.JSONDecodeError:
        return req.text
    error = result.get('error', None)
    if error is not None:
        print('Error:', error['message'])
        raise Exception(error)
        return None
    return result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import urwid

from ytsearch import ui, settings
from ytsearch.ui import page


class Interface(page.Page):

    """
    The interface for the list of playlists.

    :page.Page: The base page.
    """

    mode = 'playlist'

    def load_page(self):
        """
        Load the playlist page.

        :return: urwid.Pile: The pile widget with all of the other widgets
        """
        title = urwid.Filler(urwid.Text(('title', u'Playlists')), 'top')
        if self.parent.playlists == {}:
            keybind = settings.find_keybinding('CREATE_PLAYLIST')
            text = urwid.Text(('title', "Press '{}' to add Playlist".format(keybind)),
                              'center')
            videos = urwid.Filler(text, 'middle')
        else:
            videos = self.load_videos()
        pile = urwid.Pile([(2, title), ('weight', 1, videos)])
        self.widgets = pile
        return pile

    def load_videos(self):
        """
        Load all of the names of the playlist.

        :return: urwid.ItemList: A list of the video widgets.
        """
        videos = []
        self.results = []
        for name in dict(self.parent.playlists):
            video = ui.Video(name, None, 'playlist', None)
            self.results.append(video)
            widget = self.parent.create_video_widget(video, reuse=False)
            videos.append(widget)
        self.walker = urwid.SimpleListWalker(videos)
        self.video_list = ui.ItemList(self.parent, self.walker, 'playlist')
        return self.video_list

    def event_CREATE_PLAYLIST(self, _):
        self.parent.create_search_widget('Playlist: ', self.parent.add_playlist)
        return None

    def event_PLAY_AUDIO(self, index):
        video = self.results[index]
        playlist_name = video.name
        self.parent.load_playlist(playlist_name)
        return None

    def event_QUEUE_AUDIO(self, index):
        video = self.results[index]
        playlist_name = video.name
        self.parent.queue_playlist(playlist_name, audio=True)
        return None

    def event_QUEUE_VIDEO(self, index):
        video = self.results[index]
        playlist_name = video.name
        self.parent.queue_playlist(playlist_name, audio=False)
        return None

    def event_TOGGLE_PLAYLIST_ADD(self, index):
        if self.parent.playlist_add is None:
            video = self.results[index]
            name = video.name
            self.parent.status = 'Adding to playlist {}'.format(name)
            self.parent.playlist_add = name
        else:
            self.parent.playlist_add = None
            self.parent.status = ''
        return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import os

import urwid

from ytsearch import ui
from ytsearch.ui import page


class Interface(page.Page):

    """
    The interface class for the queued items you have saved.

    :page.Page: The base page to inherit from.
    """

    widgets = None
    results = []

    def __init__(self, parent):
        self.parent = parent

    def load_page(self):
        """
        Load the page.

        :return: urwid.Pile: The pile widget containing all other widgets
        """
        title = urwid.Filler(urwid.Text(('title', u'Queued Items')), 'top')
        videos = self.load_videos()
        pile = urwid.Pile([(2, title), ('weight', 1, videos)])
        self.widgets = pile
        return pile

    def load_videos(self):
        """
        Load all of the videos.

        :return: urwid.ItemList: The widget with all of the video lists.
        """
        videos = []
        self.results = []
        for video in self.parent.queue:
            self.results.append(video)
            widget = self.parent.create_video_widget(video)
            videos.append(widget)
        self.walker = urwid.SimpleListWalker(videos)
        self.video_list = ui.ItemList(self.parent, self.walker, 'queue')
        return self.video_list



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import os
import collections

import yaml


CONF_DIR = os.path.expanduser('~/.ytsearch')


DEFAULT = '''
keybindings:
    global:
        '1': 'PAGE search'
        '2': 'PAGE cache'
        '3': 'PAGE queue'
        '4': 'PAGE playlist'
        '?': 'PAGE settings'
        '0': 'FOCUS_PLAYER'
        'j': 'MOVE +1'
        'k': 'MOVE -1'
        'down': 'MOVE +1'
        'up': 'MOVE -1'
        'f': 'SEARCH'
        ':': 'COMMAND'
        'a': 'PLAY_AUDIO'
        'v': 'PLAY_VIDEO'
        'ZZ': 'QUIT'
        'qa': 'QUEUE_AUDIO'
        'qv': 'QUEUE_VIDEO'
        '+': 'CREATE_PLAYLIST'
        'p': 'PLAYLIST_ADD'
        'c': 'TOGGLE_CONSUME'
        'r': 'TOGGLE_REPEAT'
        '/': 'FIND_VIDEO'
        'd': 'DOWNLOAD_VIDEO'
        'n': 'NEXT_VIDEO'
        'R': 'SHUFFLE_QUEUE'
        'tp': 'SEND_KEY p'
        '=': 'SEND_KEY 0'
        '-': 'SEND_KEY 9'
        'ctrl d': 'MOVE +10'
        'ctrl u': 'MOVE -10'
        'gg': 'MOVE top'
        'G': 'MOVE bottom'
        'M': 'MOVE playing'
        'QA': 'QUEUE_ALL_AUDIO'
        'QV': 'QUEUE_ALL_VIDEO'
    playlist:
        'P': 'TOGGLE_PLAYLIST_ADD'
        'enter': 'PLAY_AUDIO'
    playlist_items:
        'delete': 'REMOVE_PLAYLIST_ITEM'
    player:
        '1': 'FOCUS_NORMAL'
    settings:
        'esc': 'PAGE cache'

quality:
    video: '1080,720,480,360'
    audio: '160k,128k,70k,48k'

player:
    command: 'mpv'
    video_args: ['--player-operation-mode=pseudo-gui']
    audio_args: ['--no-video']
    size: 5
    quit_key: 'q'

settings:
    command: ['nano']

cache: '~/videos/youtube_cache'
'''


def find_keybinding(event_name):
    """
    Find the key corresponding with an event name.
    
    :event_name: str: The event name to find.
    :return: str: The key that was found.
             None if none was found.
    """
    user_settings = load_user_settings()
    keybind = find_event_name(event_name, user_settings)
    if keybind is not None:
        return keybind
    default_settings = SETTINGS
    keybind = find_event_name(event_name, default_settings)
    return keybind


def find_event_name(event_name, default_settings, level=None):
    """
    Recursively find the event name,
    
    :event_name: str: the event name to find.
    :default_settings: dict: The settings to start on the first recursion.
    :level: dict: The level of iteration its on.
    :return: str: THe key that was found.
             None if none was found.
    """
    if level is None:
        level = default_settings
    for key, name in level.items():
        if isinstance(name, collections.Mapping):
            output = find_event_name(event_name, default_settings, name)
            if output is not None:
                return output
        if name == event_name:
            return key
    return None


def load_settings():
    """
    Loads the settings from disk.
    
    :return: dict: A dictionary of merged settings.
    """
    user_settings = load_user_settings()
    default_settings = yaml.load(DEFAULT)
    merged = merge_settings(default_settings, user_settings)
    return merged


def merge_settings(default, user):
    """
    Merge 2 dictionaries recursively.
    
    :default: dict: The base dictionary to add items to.
    :user: dict: The dictionary to add items from, if they exist.
    :return: dict: The merged contents.
    """
    for key, value in user.items():
        if (key in default and isinstance(default[key], dict)
        and isinstance(value, collections.Mapping)):
            merge_settings(default[key], value)
        else:
            default[key] = value
    return default


def load_user_settings():
    """Loads the settings.yaml file.
    
    :return: dict: The settings that were loaded.
    """
    if not os.path.exists('{}/settings.yaml'.format(CONF_DIR)):
        return {}
    with open('{}/settings.yaml'.format(CONF_DIR)) as f:
        data = f.read() 
    settings = yaml.load(data)
    if settings is None:
        return {}
    return settings


def find_different(original, new):
    output = {}
    for name, value in new.items():
        if isinstance(value, collections.Mapping) and name in original:
            result = find_different(original[name], value)
            if result != {}:
                output[name] = result
        elif original.get(name, None) != value:
            output[name] = value
    return output


def remove_missing(old, new):
    items = dict(new)
    for name, value in items.items():
        if isinstance(value, collections.Mapping):
            remove_missing(old[name], value)
        elif name not in old:
            del new[name]
    return new


def save_settings(settings):
    original = load_settings()
    difference = find_different(original, settings)
    user_settings = load_user_settings()
    merged = merge_settings(user_settings, difference)
    merged = remove_missing(settings, merged)
    dump = yaml.dump(merged, default_flow_style=False)
    with open('{}/settings.yaml'.format(CONF_DIR), 'w') as f:
        f.write(dump)

    default_settings = yaml.load(DEFAULT)
    return merge_settings(default_settings, merged)


# TODO make this load only once. Right now this executes with each import.
if not os.path.exists(CONF_DIR):
    os.mkdir(CONF_DIR)

SETTINGS = load_settings()
cache = SETTINGS.get('cache', '~/videos/youtube_cache')
if not cache.endswith('/'):
    cache += '/'

CACHE_LOCATION = os.path.expanduser(cache)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import json
import os
import random
import shlex
import subprocess
import sys
import time
import logging
import traceback
import asyncio

import urwid
import pafy
from fuzzywuzzy import fuzz, process
import youtube_dl

from ytsearch.ui import (search_results, cache_items, queue_items,
                         playlist, playlist_items)
from ytsearch.ui import settings as settings_ui
from ytsearch import settings, threads, youtube, lock


CACHE_LOCATION = settings.CACHE_LOCATION
CONF_DIR = os.path.expanduser('~/.ytsearch')
LOOP = asyncio.get_event_loop()


COLOURS = [
    ('title', 'bold', ''),
    ('standout', 'standout', ''),
    ('underline', 'underline', ''),
    ('blue', 'dark blue', ''),
    ('blue_bold', 'dark blue, bold', ''),
    ('green', 'dark green', ''),
    ('green_bold', 'dark green, bold', '')
]


SETTINGS = settings.SETTINGS
KEYBINDS = SETTINGS['keybindings']
PLAYER = SETTINGS['player']
HOOKS = SETTINGS.get('hooks', {})
LOGGER = logging.getLogger()


def reload_settings():
    """Reloads all of the settings being used by the UI."""
    global SETTINGS, KEYBINDS, PLAYER, HOOKS
    SETTINGS = settings.SETTINGS
    KEYBINDS = SETTINGS['keybindings']
    PLAYER = SETTINGS['player']
    HOOKS = SETTINGS.get('hooks', {})
    return None


def get_video_quality(video_url, qualities, audio=True):
    """
    Finds the quality to load the video / audio as.

    :video_url: str: The url of the video to load.
    :qualities: list: A list of qualities to preference, in order.
    :audio: bool: True if the video will be loaded as audio.
                  False otherwise.
    :return: str: The url of the video, according to the quality list.
             None if no url was found.
    """
    video = pafy.new(video_url)
    streams = video.audiostreams if audio else video.streams
    for stream in streams:
        if any(stream.quality.endswith(q) for q in qualities):
            return stream.url
    return None


class Hook:
    """
    A decorator that calls HOOK functions when the decorated function
    is run
    """
    def __init__(self, hook_name):
        """
        Create the hook

        :hook_name: str: The name of the hook to run.
        """
        self.hook_name = hook_name

    def __call__(self, function):
        """
        Runs when the decorator is created.

        :function: func: The function that activates the hook.
        :return: func: The function that runs the hook / function.
        """
        def run(*args, **kwargs):
            """
            Handles the funning of both the function and hook.

            :*args: list: A list of arguments to pass to the function.
            :**kwargs: dict: A dictionary of keyword args to pass.
            :return: ?: The output from the function.
            """
            run_hook, output = function(*args, **kwargs)
            if self.hook_name in HOOKS and run_hook:
                command = list(HOOKS[self.hook_name])
                for index, item in enumerate(command):
                    command[index] = item.format(*args, **kwargs)
                subprocess.run(command)
            return output
        return run 


# TODO: Hook these messages into the logger.
class VoidLogger:
    def debug(self, msg):
        pass

    def warning(self, msg):
        LOGGER.warning(msg)
        return None

    def error(self, msg):
        LOGGER.error(msg)
        return None


class EditWidget(urwid.Edit):
    """The widget that lets the user input something."""

    def __init__(self, call, *args, **kwargs):
        self.call = call
        super().__init__(*args, **kwargs)

    def keypress(self, size, key):
        if key == 'enter':
            self.call(self.text)
        if key == 'esc':
            self.call('')
        else:
            return super().keypress(size, key)


class TerminalWidget(urwid.Terminal):
    """
    The terminal widget that the user uses to interact with the
    command that is playing the video / audio
    """
    def __init__(self, parent, *args, **kwargs):
        self.parent = parent
        super().__init__(*args, **kwargs)

    def keypress(self, size, key):
        found = self.parent.unhandled_input(key, 'player')
        if not found:
            return super().keypress(size, key)
        return None


class Video:
    """
    A class to store information for each video.
    """

    selected = False
    widget = None
    terminal = None
    temporary = False
    downloading = False
    _status = ''

    def __init__(self, name, location, resource='file', cache=None):
        """
        Create a new video.
        
        :name: str: The name of the video.
        :location: str: The location of the video resource.
        :resource: str: The type of resource.
        :cache: str: A location of the cache, if any. None otherwise.
        """
        self.name = name
        self.location = location
        self._resource = resource
        self.cache = cache

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, text):
        self._status = text
        return None

    def stop(self):
        """
        Stop the video from playing by sending 'quit_key' to the process.

        :return: None
        """
        if self.terminal is None:
            return None
        kill_key = PLAYER['quit_key']
        self.terminal.respond(kill_key)
        return None
    
    def resource(self, audio=True):
        """
        Load the resource of the video.
        
        :audio: bool: True: If it should load the video as audio only.
                      False: if it should load the video too.
        :return: str: The location of the resource.
                 None: It returns None if the resource couldn't be found
        """
        if self.cache is not None:
            return self.cache
        if self._resource == 'preloaded' or self._resource == 'file':
            return self.location
        if self._resource == 'youtube':
            url = 'https://youtube.com/watch?v={}'.format(self.location)
            if audio:
                qualities = SETTINGS['quality']['audio'].split(',')
            else:
                qualities = SETTINGS['quality']['video'].split(',')
            url = get_video_quality(url, qualities, audio)
            return url
        return None

    @threads.AsThread()
    def preload(self, audio=True):
        """
        Preload the video resource. Used when you add a video to the queue
        
        :audio: bool: True: If the video should be loaded as audio.
                      False: If it should be audio and video.
        :return: None
        """
        resource = self.resource(audio)
        self._resource = 'preloaded'
        self.location = resource
        return None

    def send(self, string):
        """Send a string to the running terminal widget.
        
        :string: str: The string to send to the widget.
        :return: None
        """
        if self.terminal is not None:
            self.terminal.respond(string)
        return None

    def __eq__(self, video):
        """
        Check if a video is equal to another video.
        
        :video: Video: The video to check equality of.
        :return: bool: True: if the videos are the same.
                       False: If they are not the same.
        """
        if video is None:
            return False
        return self.location == video.location or self.name == video.name

    def __gt__(self, video):
        """
        Check if a video is greater than another video.
        This just uses the len() of the name.
        
        :video: Video: The video to compare sizes to.
        :return: bool: True: If the current video is larger than the param.
                       False: if the param is larger than this video."""
        if video is None:
            return False
        return self.name > video.name

    def __len__(self):
        """
        Return the length of the video name.
        
        :return: int: Length of the name.
        """
        return len(self.name)

    def __iter__(self):
        """
        Return the iter on the name of the video.
        
        :return: iterable: Iterable of the video name.
        """
        return iter(self.name)

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    def __getitem__(self, attr):
        if hasattr(self, attr):
            return getattr(self, attr)
        return self.name[attr]


class ItemList(urwid.ListBox):
    def __init__(self, parent, widgets, mode):
        self.parent = parent
        self.keybuffer = []
        self.mode = mode
        super().__init__(widgets)

    def find_video(self):
        """
        Find and move the cursor to the name of the video the 
        user is inputting.
        """
        # fuzzywuzzy freaks out when certain characters are passed to it.
        # As far as I can find, the warnings it gives off cannot be stopped
        # they ruin the TUI so I strip characters here.
        find = ''.join(self.keybuffer).strip('+=\'!@#$%^&*()_+"')
        self.parent.status = '/{}'.format(find)
        self.parent.loop.draw_screen()
        if find == '':
            return None
        videos = [x.name for x in self.parent.current_page.results]
        output = process.extractOne(find, videos,
                                    scorer=fuzz.partial_ratio)
        index = videos.index(output[0])
        self.set_focus(index)
        return None

    def keypress(self, size, key):
        self.keybuffer.append(key)
        mode_keys = KEYBINDS.get(self.mode, {})
        keys = dict(KEYBINDS.get('global', {}))
        keys.update(mode_keys)
        possible = []
        matches = []
        if key == 'backspace':
            self.keybuffer = self.keybuffer[:-2]
        if key == 'enter' and self.parent.find_video:
            self.parent.find_video = False
            self.parent.status = ''
            return None
        if key == 'esc' and self.parent.find_video:
            self.parent.find_video = False
            self.parent.status = ''
            return None
        if self.parent.find_video:
            self.find_video()
            return None
        if len(self.keybuffer) > 10:
            self.keybuffer = self.keybuffer[-10:]
        for key in self.keybuffer[::-1]:
            possible.insert(0, key)
            for index, key in sorted(enumerate(keys), key=lambda a: len(a),
                                     reverse=True):
                if ''.join(possible) == str(key):
                    matches.append(str(key))
        if matches != []:
            key = sorted(matches, key=lambda x: len(x))[-1]
            event = keys[key]
            _, index = self.get_focus()
            self.parent.key_event(event, index)
            self.keybuffer = []
            return None
        return None


class Interface:

    pages = {'cache': cache_items.Interface,
             'search': search_results.Interface,
             'queue': queue_items.Interface,
             'playlist': playlist.Interface,
             'playlist_items': playlist_items.Interface,
             'settings': settings_ui.Interface}
    saved_pages = {'cache': None, 'search': None, 'queue': None,
                   'playlist': None, 'playlist_items': None,
                   'settings': False}
    video_storage = []
    queue = []
    current_page = None
    page_placeholder = None
    keybuffer = []
    playing = None
    _playing_position = 0
    terminal = None
    _playlists = {}
    playlist_add = None
    state = {'consume': False, 'repeat': True}
    find_video = False
    previous_focus = 0

    def __init__(self): 
        self.player_placeholder = urwid.Filler(urwid.Text(''))
        self.input_placeholder = urwid.Filler(urwid.Text(''))
        self.load_playlists()

    def __enter__(self):
        return self

    def __exit__(self, *_):
        if self.playing is not None:
            self.playing.stop()
        self.save_playlists()
        settings.save_settings(settings.SETTINGS)
        return None

    @property
    def playlists(self):
        """
        Loads the current playlist. If the playlist is None it sends out
        an error.

        :return: dict: The current playlists.
        """
        if self._playlists is None:
            trace = ''.join(traceback.format_stack())
            logging.exception('Getter empty playlist:\n{}'.format(trace))
        return self._playlists

    @playlists.setter
    def playlists(self, value):
        """
        Sets the playlists, if you set it to None the program sends an
        error to the logfile.

        :return: None
        """
        if value is None:
            trace = ''.join(traceback.format_stack())
            logging.exception('Setter empty playlist:\n{}'.format(trace))
        self._playlists = value
        return None

    @property
    def playing_position(self):
        """
        Return the current song index. If you have random on it
        uses a random number based on the length of the queue.
        
        :return: int: The position of the song in the queue.
        """
        if (isinstance(self._playing_position, tuple)
        and self._playing_position[0]):
            return self._playing_position[1]
        return self._playing_position

    @playing_position.setter
    def playing_position(self, new):
        """
        Set the playing position.
        
        :new: int: The new position to use.
              tuple: If this is a tuple, the first argument is if
                     the position should be forced.
        :return: None
        """
        self._playing_position = new
        return None

    def load_playlists(self):
        """
        Load the playlists the user has created. If there are
        any issues loading the playlists will be set to blank.
        """
        playlist_dir = '{}/playlists.json'.format(CONF_DIR)
        lockfile = '{}/playlists.lock'.format(CONF_DIR)
        if not os.path.exists(playlist_dir):
            self.playlists = {}
            return None
        if not lock.create(lockfile):
            LOGGER.warning('Loading playlist: Cannot create lockfile.')
            return None
        with open(playlist_dir, 'r') as f:
            data = f.read()
        try:
            playlists = json.loads(data)
        except json.decoder.JSONDecodeError:
            self.playlists = {}
        else:
            self.playlists = playlists
        lock.release(lockfile)
        return None

    def save_playlists(self):
        """Save the created playlists."""
        if self.playlists is None:
            return None
        playlist_dir = '{}/playlists.json'.format(CONF_DIR)
        lockfile = '{}/playlists.lock'.format(CONF_DIR)
        if not lock.create(lockfile):
            LOGGER.warning('Saving playlist: Cannot create lockfile.')
            return None
        encoded = json.dumps(self.playlists)
        with open(playlist_dir, 'w') as f:
            f.write(encoded)
        lock.release(lockfile)
        return None

    def main(self, start_page, start_widget=None):
        """
        Create all of the widgets and run the UI.
        
        :start_page: str: The name of the page to load by default.
        :start_widget: urwid.Widget: The widget to load by default.
        :return: None
        """
        self.load_page(start_page)
        self.status_widget = urwid.Filler(urwid.Text(''))
        self.state_widget = self.create_state()
        if start_widget is not None:
            self.page_placeholder.original_widget = start_widget
        divider = urwid.Divider(u'-', 1, 1)
        self.widgets = urwid.Pile([
            ('weight', 1, self.page_placeholder),
            (0, self.input_placeholder),
            ('pack', divider),
            (0, self.player_placeholder),
            (1, urwid.Columns([self.status_widget, (5, self.state_widget)]))
            ])
        self.loop = urwid.MainLoop(self.widgets, COLOURS,
                                   unhandled_input=self.unhandled_input,
                                   event_loop=urwid.AsyncioEventLoop(loop=LOOP))
        self.loop.run()
        return None

    def create_state(self):
        """
        Create the state widget, based off the state dictionary.
        
        :return: urwid.Filler: A filler wrappring the text widget.
        """
        consume_char = settings.find_keybinding('TOGGLE_CONSUME')
        consume = consume_char if self.state['consume'] else '-'
        repeat_char = settings.find_keybinding('TOGGLE_REPEAT')
        repeat = repeat_char if self.state['repeat'] else '-'
        text = '[{}{}]'.format(consume, repeat)
        return urwid.Filler(urwid.Text(text, 'right'))

    def update_state(self):
        """Update the state widget. This will also re-draw the UI."""
        self.state_widget.original_widget = self.create_state().original_widget
        self.loop.draw_screen()
        return None

    @property
    def status(self):
        pass

    @status.setter
    def status(self, status):
        """
        Set the text of the status widget at the bottom.
        
        :status: str: The text to set on the status.
        :return: None
        """
        if (not hasattr(self, 'status_widget')
        or not isinstance(self.status_widget.original_widget, urwid.Text)):
            return None
        self.status_widget.original_widget.set_text(status)
        self.loop.draw_screen()
        return None

    @property
    def active_video(self):
        """
        Returns the active video index.

        :return: int: The index of the active video.
        """
        if self.current_page is None:
            return None
        video_list = getattr(self.current_page, 'video_list', None)
        if video_list is None:
            return None
        return video_list.get_focus()[1]

    def create_command_widget(self, call):
        """
        Loads the command input widget.

        :call: function: The function to be called when the edit widget
                         finishes its input.
        :return: None
        """
        edit = EditWidget(call, ':')
        self.status_widget.original_widget = edit
        self.previous_focus = self.widgets.focus_position
        self.widgets.focus_position = 4
        self.loop.draw_screen()
        return None

    def destroy_command_widget(self):
        """Destroys the command input widget."""
        self.status_widget.original_widget = urwid.Text('')
        self.widgets.focus_position = self.previous_focus
        self.loop.draw_screen()
        return None

    def command_event(self, command):
        """
        Called when you input a command.

        :command: str: The command + params to be called.
        :return: None
        """
        video_index = self.active_video
        self.destroy_command_widget()
        command, *params = command.split(' ', 1)
        params.insert(0, video_index)
        command = command.lstrip(':')
        command_name = 'event_{}'.format(command.upper())
        call = (getattr(self.current_page, command_name, None) or
                getattr(self, command_name, None) or
                getattr(self.terminal, command_name, None))
        if call is not None:
            try:
                call(*params)
            except Exception as e:
                self.status = str(e)
        return None

    def create_search_widget(self, text, call):
        """
        Create the search widget for the user to input some text
        
        :text: str: The text to display next to the input.
        :call: func: The function to call when the user hits enter.
        :return: None
        """
        title = urwid.Text(text)
        edit = EditWidget(call)
        column = urwid.Columns([(len(text), title), edit])
        padding = urwid.Padding(column, left=2, right=2)
        filler = urwid.Filler(padding, 'middle')
        attrmap = urwid.AttrMap(filler, 'standout', None)
        self.widgets.contents[1] = (attrmap, ('given', 3))
        self.widgets.focus_position = 1
        self.loop.draw_screen()
        return None

    def destroy_search_widget(self):
        """Destroys the search widget and redraws the UI."""
        widget = urwid.Filler(urwid.Text(''))
        self.widgets.contents[1] = (widget, ('given', 0))
        self.widgets.focus_position = 0
        self.loop.draw_screen()
        return None

    @threads.AsThread()
    def search(self, video_name):
        """
        Search youtube for a videos, it will update the search
        results page if its active or in saved pages.
        
        :video_name: str: The name of the video to search for.
        :return: None
        """
        self.destroy_search_widget()
        if video_name == '':
            return None
        results = youtube.search(video_name)
        if results is None:
            return None
        if isinstance(self.current_page, search_results.Interface):
            self.current_page.results = results
            self.current_page.load_page()
            self.load_page('search')
            self.loop.draw_screen()
        elif self.saved_pages['search'] is not None:
            self.saved_pages['search'].results = results
            self.saved_pages['search'].load_page()
        else:
            search_page = self.pages['search'](self)
            search_page.results = results
            search_page.load_page()
            self.saved_pages['search'] = search_page
        return None

    def load_page(self, name, switch_focus=False):
        """
        Load one of the pages. Optionally switch focus to the page
        section of the UI.
        
        :name: str: The name of the page to load.
        :switch_focus: bool: True if the UI focus should be switched
                                  to the page section.
                             False otherwise
        :return: urwid.WidgetPlaceholder: The new page placeholder.
        """
        self.status = ''
        saved = self.saved_pages.get(name, None)
        if saved is not None and saved is not False:
            page = saved
            widgets = page.widgets
        elif not saved:
            page = self.pages[name](self)
            widgets = page.load_page()
            if self.saved_pages[name] is not False:
                self.saved_pages[name] = page
        self.current_page = page
        if self.page_placeholder is None:
            self.page_placeholder = widgets
        else:
            self.page_placeholder = widgets
            self.widgets.contents[0] = (self.page_placeholder, ('weight', 1))
            if switch_focus:
                self.widgets.focus_position = 0
        self.update_videos()
        return self.page_placeholder

    def update_videos(self):
        """
        Update the videos that have been placed in the video storage.
        Basically these are any videos that have been created.
        """
        for video in self.video_storage:
            if video in self.current_page.results:
                index = self.current_page.results.index(video)
                new_widget = self.create_video_widget(video)
                self.current_page.walker[index] = new_widget
        return None

    def key_event(self, event, index):
        """
        Called when the input found a key combo that has been bound.
        
        :event: str: The name of the event that was found.
        :index: int: The index of the cursor on the video list.
        :return: None
        """
        name, *params = event.split(' ')
        params.insert(0, index)
        event_name = 'event_{}'.format(name)
        call = (getattr(self.current_page, event_name, None) or
                getattr(self, event_name, None) or
                getattr(self.terminal, event_name, None))
        if call is not None:
            call(*params)
        return None

    def create_video_widget(self, video, reuse=True):
        """
        Create the widget that represents one of the videos.
        
        :video: Video: The video to create a widget for.
        :reuse: bool: True if this should re-use a video from the
                           video storage.
                      False otherwise.
        :return: urwid.AttrMap: The attrmap surrounding the video widget.
        """
        if video in self.video_storage and reuse:
            video = self.video_storage[self.video_storage.index(video)]
        elif reuse:
            self.video_storage.append(video)
        focusmap = {'': 'title'}
        video_colour = ''
        status = video.status
        name = video.name
        if self.playing == video:
            video_colour = 'green'
            focusmap = {'green': 'green_bold'}
        elif video.cache is not None:
            video_colour = 'blue'
            focusmap = {'blue': 'blue_bold'}
        widget = urwid.Text((video_colour, name))
        status_widget = urwid.Text((video_colour, status), 'right')
        column = urwid.Columns([('pack', widget), status_widget])
        output = urwid.AttrMap(column, None, focusmap)
        video.widget = output
        return output

    def load_playlist(self, playlist_name):
        """
        Load a playlist and populates the playlist page.
        
        :playlist_name: str: The name of the playlist to load.
        :return: None
        """
        page = self.pages['playlist_items'](self)
        page.playlist_items = list(self.playlists[playlist_name])
        page.description = playlist_name
        page.playlist_name = playlist_name
        page.load_page()
        self.saved_pages['playlist_items'] = page
        self.load_page('playlist_items')
        return None

    @Hook('PLAY')
    def play(self, video, audio=False):
        """
        Play a video, optionally as audio.
        
        :video: Video: The video to play.
        :audio: bool: True if you want to only play audio.
                      False otherwise.
        :return: bool: True if the PLAY hook should run.
                       False otherwise.
                 bool: True if the cursor should move one down.
                       False otherwise.
        """
        video.audio = audio
        if video == self.playing:
            video.stop()
            return False, False
        if video not in self.queue:
            self.queue.append(video)
            self.playing_position += 1
        if self.playing is not None:
            if video in self.queue:
                self.playing_position = (True, self.queue.index(video))
            else:
                self.queue.insert(0, video)
                self.playing_position = 0
            self.playing.stop() 
            return False, False
        self.playing = video

        LOOP.call_later(0, self.run_command, video, audio)
        return True, True

    def play_finish(self, video, audio=False, force=False):
        """
        Called when the video finishes playing.
        
        :video: Video: The video that was playing.
        :audio: bool: True if the video was played as audio.
                      False otherwise.
        force: bool: True if it shold force the video to stop.
                     False otherwise.
        """
        if self.state['consume'] and video in self.queue and not force:
            index = self.queue.index(video)
            self.playing_position -= 1
            del self.queue[index]
            self.update_queue_page()
        self.set_video_status(video, '')
        self.playing = None
        if self.queue != []:
            if self.playing_position >= len(self.queue):
                if self.state['repeat'] and not force:
                    self.playing_position = 0
                else:
                    self.destroy_terminal_widget()
                    self.clear_queue()
                    self.update_queue_page()
                    return None
            next_video = self.queue[self.playing_position]
            self.playing_position += 1
            self.play(next_video, next_video.audio)
            self.update_queue(update_page=False)
        else:
            self.destroy_terminal_widget()
        return None

    def clear_queue(self):
        """Clear all of the videos in the queue."""
        for video in self.queue:
            self.set_video_status(video, '')
        self.queue = []
        return None

    def queue_add(self, video, redraw=True):
        """
        Add a video to the queue. It will remove items if they
        are already in the queue.
        
        :video: Video: The video to add to the queue.
        :redraw: bool: True if the UI should be reloaded with the addition
                       False otherwise.
        :return: None
        """
        if self.playing is None:
            self.play(video, video.audio)
            return None
        if video == self.playing:
            return False
        if video in self.queue:
            index = self.queue.index(video)
            del self.queue[index]
            self.set_video_status(video, '')
        else:
            self.queue.append(video)
            status = 'Queue #{}'.format(len(self.queue))
            self.set_video_status(video, status)
            index = len(self.queue)
        if redraw:
            self.loop.draw_screen()
            self.update_queue(index-2)
        return None

    def queue_playlist(self, playlist_name, audio=True):
        """
        Add an entire playlist to the queue. This may be unneeded
        with the addition of QUEUE_ALL

        :playlist_name: str: The name of the playlist.
        :audio: bool: True if the videos should be played as audio.
                      False otherwise.
        :return: None
        """
        videos = []
        playlist_data = self.playlists[playlist_name]
        for data in playlist_data:
            video = Video(data['name'], data['location'], data['resource'],
                          data['cache'])
            if video in self.video_storage:
                index = self.video_storage.index(video)
                video = self.video_storage[index]
            video.audio = audio
            self.video_storage.append(video)
            self.queue_add(video, redraw=False)
        self.update_queue()
        return None

    def update_queue_page(self):
        """
        Update the queue page then redraw the UI.
        
        :return: None
        """
        if isinstance(self.current_page, queue_items.Interface):
            self.current_page.load_page()
            self.load_page('queue', False)
            self.loop.draw_screen()
            return None
        if self.saved_pages['queue'] is None:
            return None
        self.saved_pages['queue'].load_page()
        return None

    def update_queue(self, start=0, update_page=True):
        """
        Update the queue, re-setting the status's and redrawing the page.

        :start: int: The video index to start at.
        :update_page: bool: True if the page should be redrawn.
                            False otherwise.
        :return: None
        """
        playing_index = None
        if self.playing in self.queue:
            playing_index = self.queue.index(self.playing)
        for index, video in enumerate(self.queue[start:]):
            if self.playing == video:
                continue
            video_index = index - playing_index
            if video_index < 0:
                video_index = len(self.queue) - abs(video_index)
            self.set_video_status(video, 'Queue #{}'.format(
                                  video_index + start + 1))
        if update_page:
            self.update_queue_page()
        return None

    #TODO: Probably should re-name this to update_cache_page
    def update_cache(self):
        """Update the cache page."""
        self.update_videos()
        if isinstance(self.current_page, cache_items.Interface):
            self.current_page.load_page()
            self.load_page('cache')
            self.loop.draw_screen()
        elif self.saved_pages['cache'] is not None:
            self.saved_pages['cache'].load_page()
        return None

    def set_video_status(self, video, status):
        """
        Set the text next to a video. It will also re-draw the UI

        :video: Video: The video to change the status of.
        :status: str: The new status to change to.
        :return: None
        """
        video.status = status
        self.current_page.update_video(video)
        if video in self.video_storage:
            index = self.video_storage.index(video)
            self.video_storage[index] = video
        return None

    @threads.AsThread()
    def run_command(self, video, audio=False):
        """
        Runs the user specified command for running the video / audio.

        :video: Video: The video to play.
        :audio: bool: True if it should be played as audio.
                      False otherwise.
        :return: None
        """
        command = PLAYER['command']
        arg_settings = 'audio_args' if audio else 'video_args'
        arguments = PLAYER[arg_settings]
        previous = video.status
        self.set_video_status(video, 'Loading')
        self.loop.draw_screen()
        resource = video.resource(audio)
        if resource is None:
            return None
        self.set_video_status(video, 'Playing')
        full_command = [command] + arguments + [resource]
        self.create_terminal_widget(video, full_command, audio)
        return None

    @threads.AsThread()
    def download_video(self, video):
        """
        Starts a download of the video. Thanks to youtube-dl

        :video: Video: The video to download.
        :return: None
        """
        if video.downloading or video._resource != 'youtube':
            return None
        video.download = True
        video_id = video.location
        video_name = video.name
        if os.path.exists(video_id):
            return None
        self.set_video_status(video, 'Pending...')
        url = 'https://youtube.com/watch?v={}'.format(video_id)
        options = {
            'logger': VoidLogger(),
            'progress_hooks': [lambda i: self.download_handler(video, i)],
            'outtmpl': '{}{}.%(ext)s'.format(CACHE_LOCATION, video_name)
            }
        ydl = youtube_dl.YoutubeDL(options)
        ydl.download([url])
        cached = {os.path.splitext(v)[0]: v for v in os.listdir(CACHE_LOCATION)}
        video.cache = cached.get(video_name, None)
        self.update_cache()
        self.download_finished(video)
        self.loop.draw_screen()
        return None

    def download_handler(self, video, info):
        """
        Handles information given to the program from youtube-dl

        :video: Video: The video that is being downloaded
        :info: dict: Information from youtube-dl
        :return: None
        """
        if info['status'] == 'downloading':
            percentage = info['downloaded_bytes'] / info['total_bytes'] * 100
            status = '{}%'.format(round(percentage))
            self.set_video_status(video, status)
        return None

    @Hook('DOWNLOAD_FINISHED')
    def download_finished(self, video):
        """
        Called when the download finishes.

        :video: Video: The video that was downloading.
        :return: bool: True if it should run the DOWNLOAD_FINISHED hook
                       False otherwise
                 None
        """
        self.set_video_status(video, 'Finished')
        return True, None

    def add_playlist(self, playlist_name):
        """
        Add a playlist to the playlist list (yo dawg I heard you like lists)

        :playlist_name: str: The name of the playlist to add.
        :return: None
        """
        self.destroy_search_widget()
        if playlist_name == '':
            return None
        self.playlists[playlist_name] = []
        if isinstance(self.current_page, playlist.Interface):
            self.current_page.load_page()
            self.load_page('playlist')
            self.loop.draw_screen()
        elif self.saved_pages['playlist'] is not None:
            self.saved_pages['playlist'].load_page()
        return None

    def create_terminal_widget(self, video, command, audio=False):
        """
        Creates and runs the terminal widget.

        :video: Video: The video to play.
        :command: list: The command that will be run in the terminal
        :audio: bool: True if it should be run as audio,
                      False otherwise
        :return: None
        """
        finish = lambda *args: self.play_finish(video, audio=audio)
        self.terminal = TerminalWidget(self, command, main_loop=self.loop)
        self.widgets.contents[3] = (self.terminal, ('given', PLAYER['size']))
        video.terminal = self.terminal
        urwid.connect_signal(self.terminal, 'closed', finish)
        self.loop.draw_screen()
        return None

    def destroy_terminal_widget(self):
        """Destroys the terminal widget"""
        self.widgets.contents[3] = (urwid.Filler(urwid.Text('')), ('given', 0))
        self.widgets.focus_position = 0
        return None

    def restore_terminal_widget(self):
        """Unhides the terminal widget."""
        if self.terminal is None:
            return None
        self.widgets.contents[3] = (self.terminal, ('given', PLAYER['size']))
        return None

    def unhandled_input(self, key, mode='global'):
        """
        Gets called when the main UI recieves input that was not
        sent to any of the page widgets.

        :key: str: The key that was sent.
        :mode: str: The keybindings to use.
        :return: bool: True if there was a keybinding found.
                       False otherwise.
        """
        if isinstance(key, tuple):
            return False
        self.keybuffer.append(key)
        keys = KEYBINDS.get(mode, {})
        possible = []
        matches = []
        if len(self.keybuffer) > 10:
            self.keybuffer = self.keybuffer[-10:]
        if key == 'backspace':
            self.keybuffer = self.keybuffer[:-2]
        for key in self.keybuffer[::-1]:
            possible.insert(0, key)
            for index, key in sorted(enumerate(keys), key=lambda a: len(a),
                                     reverse=True):
                if ''.join(possible) == str(key):
                    matches.append(str(key))
        if matches != []:
            key = sorted(matches, key=lambda x: len(x))[-1]
            event = keys[key]
            self.key_event(event, 0)
            self.keybuffer = []
            return True
        return False

    def event_PAGE(self, _, page_name):
        self.load_page(page_name)
        self.loop.draw_screen()
        return None

    def event_SEND_KEY(self, _, key):
        if self.playing is not None:
            self.playing.terminal.respond(key)
        return None

    def event_FOCUS_PLAYER(self, _):
        if self.playing is not None:
            self.widgets.focus_position = 3
        return None

    def event_FOCUS_NORMAL(self, _):
        self.widgets.focus_position = 0
        return None

    def event_SEARCH(self, _):
        self.create_search_widget('search: ', self.search)
        return None

    def event_QUIT(self, _):
        raise urwid.ExitMainLoop
        return None

    def event_CREATE_PLAYLIST(self, _):
        self.create_search_widget('Playlist: ', self.add_playlist)
        return None

    def event_TOGGLE_CONSUME(self, _):
        self.state['consume'] = not self.state['consume']
        self.update_state()
        return None

    def event_TOGGLE_REPEAT(self, _):
        self.state['repeat'] = not self.state['repeat']
        self.update_state()
        return None

    def event_SHUFFLE_QUEUE(self, _):
        for index, video in enumerate(self.queue):
            rand_index = random.randint(0, len(self.queue) - 1)
            self.queue[index], self.queue[rand_index] = (
                    self.queue[rand_index], self.queue[index])

        if self.playing in self.queue:
            index = self.queue.index(self.playing)
            self.queue[0], self.queue[index] = self.queue[index], self.queue[0]
            self.playing_position = 1
        self.update_queue()
        return None

    def event_FIND_VIDEO(self, _):
        self.find_video = True
        self.status = '/'
        return None
    
    def event_COMMAND(self, _):
        self.create_command_widget(self.command_event)
        return None
        if matches != []:
            key = sorted(matches, key=lambda x: len(x))[-1]
            event = keys[key]
            self.key_event(event, 0)
            self.keybuffer = []
            return True
        return False

    def event_PAGE(self, _, page_name):
        self.load_page(page_name)
        self.loop.draw_screen()
        return None

    def event_SEND_KEY(self, _, key):
        if self.playing is not None:
            self.playing.terminal.respond(key)
        return None

    def event_FOCUS_PLAYER(self, _):
        if self.playing is not None:
            self.widgets.focus_position = 3
        return None

    def event_FOCUS_NORMAL(self, _):
        self.widgets.focus_position = 0
        return None

    def event_SEARCH(self, _):
        self.create_search_widget('search: ', self.search)
        return None

    def event_QUIT(self, _):
        raise urwid.ExitMainLoop
        return None

    def event_CREATE_PLAYLIST(self, _):
        self.create_search_widget('Playlist: ', self.add_playlist)
        return None

    def event_TOGGLE_CONSUME(self, _):
        self.state['consume'] = not self.state['consume']
        self.update_state()
        return None

    def event_TOGGLE_REPEAT(self, _):
        self.state['repeat'] = not self.state['repeat']
        self.update_state()
        return None

    def event_SHUFFLE_QUEUE(self, _):
        for index, video in enumerate(self.queue):
            rand_index = random.randint(0, len(self.queue) - 1)
            self.queue[index], self.queue[rand_index] = (
                    self.queue[rand_index], self.queue[index])

        if self.playing in self.queue:
            index = self.queue.index(self.playing)
            self.queue[0], self.queue[index] = self.queue[index], self.queue[0]
            self.playing_position = 1
        self.update_queue()
        return None

    def event_FIND_VIDEO(self, _):
        self.find_video = True
        self.status = '/'
        return None
    
    def event_COMMAND(self, _):
        self.create_command_widget(self.command_event)
        return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import os
import time


def create(lockfile):
    """
    Create the lockfile. This will also attempt to aquire it to make
    sure it has the rights.

    :lockfile: str: The path of the lockfile to use.
    :return: bool: True if this instance has ownership, False otherwise.
    """
    if not aquire(lockfile):
        return False

    with open(lockfile, 'w') as f:
        f.write('This file is protected!')
    return True


def aquire(lockfile, timeout=1, attempts=10):
    """
    Attempt to aquire ownership over the lockfile. ie: If it exists
    for too long, this instance will NOT have ownership.

    :lockfile: str: The path of the lockfile to use.
    :timeout: int: The time to sleep between each attempt.
    :attempts: int: The number of attempts to try to get ownership.
    :return: bool: True if this instance has ownership. False otherwise.
    """
    for i in range(attempts):
        if os.path.exists(lockfile):
            time.sleep(timeout)
        else:
            return True
    return False


def release(lockfile):
    """Remove the lockfile. This is assumed to have ownership.
    You should not be calling this when the instance does not have ownership.
    
    :lockfile: str: The path of the lockfile to use.
    :return: None
    """
    os.remove(lockfile)
    return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-


import argparse
import os
import time
import json
import subprocess
import logging
import re

import urwid
import yaml
import requests

from ytsearch import youtube, ui, settings


CONF_DIR = settings.CONF_DIR
CACHE_LOCATION = settings.CACHE_LOCATION

LOG = settings.SETTINGS.get('LOG', False)
logging.basicConfig(filename='{}/output.log'.format(CONF_DIR),
                    level=logging.DEBUG)
logger = logging.getLogger()
logger.disabled = not LOG


def mkdir_recursive(directory):
    """Iterates through the directories and makes them if they don't exist.
    The same as mkdir -p
    
    :directory: The fullpath to make.
    :returns: None

    """
    split = directory.split('/')
    for index, item in enumerate(split):
        path = '/'.join(split[:index])
        if path == '':
            continue
        if not os.path.exists(path):
            os.mkdir(path)
    return None


def command_cache(_, *params):
    '''Show the cache UI'''
    with ui.Interface() as interface:
        interface.main('cache')
    return None


def command_search(_, *params):
    '''Load videos from a search result.'''
    if len(params) == 0:
        print('You must specify a search term.')
        return None
    video_name = ' '.join(params)
    results = youtube.search(video_name)
    with ui.Interface() as interface:
        widget = ui.search_results.Interface(interface)
        widget.description = video_name
        widget.results = results
        interface.saved_pages['search'] = widget
        interface.current_page = widget
        interface.main('search', widget.load_page())
    return None


def command_import(_, url, *params):
    """
    Imports a youtube playlist.

    :url: str: The url to load.
    :params: tuple: Unused.
    :return: None
    """
    print('Retrieving playlist')
    videos, info = youtube.get_playlist(url)
    playlist_name = info['snippet']['title']
    channel_name = info['snippet']['channelTitle']
    title = '{} - {}'.format(channel_name, playlist_name)
    playlist = {title: []}
    print('Parsing videos from {}'.format(title))
    for video in videos:
        data = {'name': video['snippet']['title'], 
                'location': video['contentDetails']['videoId'],
                'resource': 'youtube', 'cache': None}
        playlist[title].append(data)
    playlist_file = '{}/playlists.json'.format(CONF_DIR)
    if os.path.exists(playlist_file):
        with open(playlist_file, 'r') as f:
            playlists = json.loads(f.read())
    else:
        playlists = {}
    playlists.update(playlist)
    with open(playlist_file, 'w') as f:
        f.write(json.dumps(playlists))
    print('Saved playlist {}'.format(title))
    return None


def command_channel(_, *username):
    username = ' '.join(username)
    results = youtube.raw_search(username)
    videos = results['items']
    channel_id = videos[0]['snippet']['channelId']
    channel_name = videos[0]['snippet']['channelTitle']
    proceed = input('Load videos from "{}" [y/n] '.format(channel_name))
    if proceed.lower() not in ['yes', 'y']:
        return None

    channel_info = youtube.request('channels', part='contentDetails',
                                   id=channel_id, key=youtube.KEY)
    upload_playlist = channel_info['items'][0]['contentDetails']['relatedPlaylists']['uploads']
    data, name = youtube.get_playlist(upload_playlist)
    videos = []
    for info in data:
        video = ui.Video(info['snippet']['title'],
                         info['contentDetails']['videoId'],
                         'youtube', None)
        videos.append(video)

    with ui.Interface() as interface:
        widget = ui.search_results.Interface(interface)
        widget.description = channel_name
        widget.results = videos
        interface.saved_pages['search'] = widget
        interface.current_page = widget
        interface.main('search', widget.load_page())
    return None


def command_url(_, *urls):
    videos = []
    for url in urls:
        match = re.search('watch\?v=(.*?)$', url)
        if match is None:
            continue
        info = youtube.request('videos', part='snippet', id=match.group(1),
                               key=youtube.KEY)
        video_info = info['items'][0]['snippet']
        video = ui.Video(video_info['title'], match.group(1), 'youtube', None)
        videos.append(video)

    with ui.Interface() as interface:
        widget = ui.search_results.Interface(interface)
        widget.description = '{} videos'.format(len(videos))
        widget.results = videos
        interface.saved_pages['search'] = widget
        interface.current_page = widget
        interface.main('search', widget.load_page())
    return None


def check_default_files():
    if not os.path.exists(CACHE_LOCATION):
        mkdir_recursive(CACHE_LOCATION)
    return None


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('command', help='The command to run.', nargs='?')
    parser.add_argument('args', nargs='*',
                        help='The arguments to pass to the command.')
    args = parser.parse_args()
    commands = {c[8:]:globals()[c] for c in globals()
                if c.startswith('command_')}
    command = commands.get(args.command, None)
    if command is None:
        command = command_cache
        arguments = [args.command] + args.args
    else:
        if args.args is None:
            arguments = []
        else:
            arguments = args.args
    command(args, *arguments) 
    return None


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import os

import urwid
import yaml

from ytsearch import ui, settings
from ytsearch.ui import page


class Interface(page.Page):

    mode = 'settings'

    def load_page(self):
        self.parent.destroy_terminal_widget()
        template = settings.SETTINGS
        setting_file = '{}/template.yaml'.format(ui.CONF_DIR)
        with open(setting_file, 'w') as f:
            f.write(yaml.dump(template, default_flow_style=False))
        command = list(settings.SETTINGS['settings']['command'])
        command.append(setting_file)
        self.widgets = urwid.Terminal(command, main_loop=self.parent.loop)
        urwid.connect_signal(self.widgets, 'closed', self.finished)
        return self.widgets

    def finished(self, *_):
        setting_file = '{}/template.yaml'.format(ui.CONF_DIR)
        with open(setting_file, 'r') as f:
            data = yaml.load(f.read())
        os.remove(setting_file)
        settings.SETTINGS = settings.save_settings(data)
        ui.reload_settings()
        self.parent.load_page('cache')
        self.parent.update_state()
        self.parent.restore_terminal_widget()
        return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import threading


class AsThread:
    def __init__(self, daemon=True):
        self.daemon = daemon

    def __call__(self, function):
        def run(*args, **kwargs):
            thread = threading.Thread(target=function, args=args, kwargs=kwargs)
            thread.daemon = self.daemon
            thread.start()
            return None
        return run



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from setuptools import setup


def read(filename):
    data = None
    with open(filename, 'r') as f:
        data = f.read()
    return data


setup(
    name='ytsearch',
    version='0.2.7',
    description='A program to search and diplay youtube videos.',
    author='Steven J. Core',
    license='GPL3.0',
    packages=['ytsearch', 'ytsearch.ui'],
    zip_safe=False,
    include_package_data=True,
    install_requires=[
        'youtube-dl',
        'urwid',
        'requests',
        'pyyaml',
        'pafy',
        'clipboard',
        'fuzzywuzzy',
        'python-Levenshtein'
        ],
    entry_points={
        'console_scripts': [
            'ytsearch = ytsearch.program:main',
            ]
        })



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from ytsearch import settings

import urwid


class Page:

    """
    The base class for all of the page widget sections.
    All of the keypress actions that are the same in all of the pages
    are defined here
    """

    video_list = None
    results = []
    title = ''
    widgets = None
    keybuffer = []

    def __init__(self, parent):
        self.parent = parent

    def load_page(self):
        """
        The default action for loading the page, which returns a blank text.

        :return: urwid.Filler: The filler around the empty text
        """
        return urwid.Filler(urwid.Text(''))

    def update_video(self, video):
        """
        Find and update a certain video widget.

        :video: ytsearch.Video: The video to find and update.
        :return: None
        """
        if video not in self.results:
            return None
        index = self.results.index(video)
        self.walker[index] = self.parent.create_video_widget(video)
        return None

    def event_MOVE(self, index, count):
        if self.video_list is None or index is None:
            return None
        if (count == 'playing' and self.parent.playing is not None
        and self.parent.playing in self.results):
            new_position = self.results.index(self.parent.playing)
            direction = 'above' if index < new_position else 'below'
        elif count == 'bottom':
            new_position = len(self.video_list.body) - 1
            direction = 'above'
        elif count == 'top':
            new_position = 0
            direction = 'below'
        else:
            try:
                count = int(count)
            except ValueError:
                count = 0
            new_position = index + count
            direction = 'above' if count > 0 else 'below'
        length = len(self.video_list.body)
        if direction == 'above' and new_position > length - 1:
            new_position = length - 1
        if direction == 'below' and new_position < 0:
            new_position = 0
        self.video_list.set_focus(new_position, direction)
        return None

    def event_PLAY_AUDIO(self, index):
        if index is None:
            return None
        video = self.results[index]
        self.parent.play(video, audio=True)
        self.event_MOVE(index, 1)
        return None

    def event_PLAY_VIDEO(self, index):
        if index is None:
            return None
        video = self.results[index]
        self.parent.play(video, audio=False)
        self.event_MOVE(index, 1)
        return None

    def event_QUEUE_AUDIO(self, index):
        if index is None:
            return None
        video = self.results[index]
        video.audio = True
        self.parent.queue_add(video)
        self.event_MOVE(index, 1)
        return None

    def event_QUEUE_VIDEO(self, index):
        if index is None:
            return None

        video = self.results[index]
        video.audio = False
        self.parent.queue_add(video)
        self.event_MOVE(index, 1)
        return None

    def event_PLAYLIST_ADD(self, index):
        if index is None or self.parent.playlist_add is None:
            return None
        video = self.results[index]
        playlist_name = self.parent.playlist_add
        data = {'name': video.name, 'location': video.location, 'resource':
                video._resource, 'cache': video.cache}
        if data in self.parent.playlists[playlist_name]:
            return None
        self.parent.playlists[playlist_name].append(data)
        self.parent.set_video_status(video, 'Added')
        return None

    def event_DOWNLOAD_VIDEO(self, index):
        if index is None:
            return None
        video = self.results[index]
        self.parent.download_video(video)
        return None

    def event_NEXT_VIDEO(self, _):
        if self.parent.playing is not None:
            self.parent.playing.stop()
        return None

    def event_QUEUE_ALL_AUDIO(self, index):
        for video in self.results[index:]:
            video.audio = True
            self.parent.queue_add(video, False)
        self.parent.update_queue()
        return None
    
    def event_QUEUE_ALL_VIDEO(self, index):
        for video in self.results[index:]:
            video.audio = False
            self.parent.queue_add(video, False)
        self.parent.update_queue()
        return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import asyncio

import urwid

from ytsearch import ui, settings, threads
from ytsearch.ui import page


LOOP = asyncio.get_event_loop()


class Interface(page.Page):

    """
    The interface for the playlist items.

    :page.Page: The base page to load.
    """
    mode = 'playlist_items'
    playlist_name = None
    videos = []

    def load_page(self):
        """
        Load the widgets of this page.

        :return: urwid.Pile: The pile with all of the widgets in there.
        """
        if self.description != '':
            description = " - '{}'".format(self.description)
        else:
            description = ''
        title = urwid.Filler(urwid.Text([('title', 'Playlist'),
                             '{}'.format(description)]), 'top')
        header = urwid.Columns([(title)])
        self.walker = urwid.SimpleListWalker([])
        self.video_list = ui.ItemList(self.parent, self.walker, 'playlist_items')
        pile = urwid.Pile([(2, header), ('weight', 1, self.video_list)])
        self.widgets = pile

        LOOP.call_later(0, self.load_videos)
        return pile

    @threads.AsThread()
    def load_videos(self):
        """
        Load all of the videos from the playlist storage.

        :return: urwid.ItemList: The list of widgets to load.
        """
        self.results = []
        for data in self.playlist_items:
            video = ui.Video(data['name'], data['location'],
                          data['resource'], data['cache'])
            widget = self.parent.create_video_widget(video)
            self.results.append(video)
            self.walker.append(widget)
            self.parent.loop.draw_screen()
        self.parent.status = 'Loaded {} videos'.format(len(self.results))
        return None

    def event_REMOVE_PLAYLIST_ITEM(self, index):
        if self.video_list is None:
            return None
        video = self.results[index]
        name = self.playlist_name
        data = {'name': video.name, 'location': video.location,
                'resource': video._resource, 'cache': video.cache}
        if data in list(self.parent.playlists[name]):
            playlist_index = self.parent.playlists[name].index(data)
            del self.parent.playlists[name][playlist_index]
            self.parent.load_playlist(name)
        return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import urwid


from ytsearch import settings, ui
from ytsearch.ui import page


class Interface(page.Page):

    results = []
    title = 'Search Results'
    description = ''
    widgets = None
    video_list = None

    def __init__(self, parent):
        self.parent = parent
    
    def load_page(self):
        if self.description != '':
            description = " - '{}'".format(self.description)
        else:
            description = ''
        title = urwid.Filler(urwid.Text([('title', self.title),
                             '{}'.format(description)]), 'top')
        header = urwid.Columns([(title)])
        if self.results == []:
            keybind = settings.find_keybinding('SEARCH')
            text = urwid.Text(('title', "Press '{}' to Search".format(keybind)),
                              'center')
            body = urwid.Filler(text, 'middle')
        else:
            videos = [self.parent.create_video_widget(v) for v in self.results]
            self.walker = urwid.SimpleListWalker(videos)
            self.video_list = ui.ItemList(self.parent, self.walker, 'search')
            body = self.video_list
        pile = urwid.Pile([(2, header), body])
        self.widgets = pile
        return pile
